unit cnocExpressionEvaluator;

{$mode objfpc}{$H+}

interface

uses
  Classes,
  Math,
  SysUtils,
  Generics.Collections,
  Contnrs,
  PParser,
  PScanner,
  PasTree;

type
  TEvaluationValueKind = (evkNone, evkString, evkBoolean, evkInteger, evkCardinal, evkFloat);

  { TEvaluationValue }

  TEvaluationValue = class
  private
  protected
    FKind: TEvaluationValueKind;
    function GetAsExpressionString: string; virtual;
  public
    function TryGetAsString(out Str: string; out ErrMessage: string): Boolean; virtual;
    function TryGetAsInteger(out Int: Int64; out ErrMessage: string): Boolean; virtual;
    function TryGetAsCardinal(out Cardinal: QWord; out ErrMessage: string): Boolean; virtual;
    function TryGetAsFloat(out Float: ValReal; out ErrMessage: string): Boolean; virtual;
    function TryGetAsBoolean(out Bool: Boolean; out ErrMessage: string): Boolean; virtual;
    function Clone: TEvaluationValue; virtual;
    property Kind: TEvaluationValueKind read FKind;
    property AsExpressionString: string read GetAsExpressionString;
  end;

  { TIntegerEvaluationValue }

  TIntegerEvaluationValue = class(TEvaluationValue)
  protected
    FValue: Int64;
    function GetAsExpressionString: string; override;
  public
    constructor Create(AValue: Int64);
    function Clone: TEvaluationValue; override;
    function TryGetAsInteger(out Int: Int64; out ErrMessage: string): Boolean; override;
    function TryGetAsFloat(out Float: ValReal; out ErrMessage: string): Boolean; override;
  end;

  { TStringEvaluationValue }

  TStringEvaluationValue = class(TEvaluationValue)
  protected
    FValue: string;
    function GetAsExpressionString: string; override;
  public
    constructor Create(AValue: string);
    function Clone: TEvaluationValue; override;
    function TryGetAsString(out Str: string; out ErrMessage: string): Boolean; override;
  end;

  { TBooleanEvaluationValue }

  TBooleanEvaluationValue = class(TEvaluationValue)
  protected
    FValue: Boolean;
    function GetAsExpressionString: string; override;
  public
    constructor Create(AValue: Boolean);
    function Clone: TEvaluationValue; override;
    function TryGetAsBoolean(out Bool: Boolean; out ErrMessage: string): Boolean; override;
  end;

  { TCardinalEvaluationValue }

  TCardinalEvaluationValue = class(TEvaluationValue)
  protected
    FValue: QWord;
    function GetAsExpressionString: string; override;
  public
    constructor Create(AValue: QWord);
    function Clone: TEvaluationValue; override;
    function TryGetAsCardinal(out Cardinal: QWord; out ErrMessage: string): Boolean; override;
    function TryGetAsInteger(out Int: Int64; out ErrMessage: string): Boolean; override;
    function TryGetAsFloat(out Float: ValReal; out ErrMessage: string): Boolean; override;
  end;

  { TFloatEvaluationValue }

  TFloatEvaluationValue = class(TEvaluationValue)
  protected
    FValue: ValReal;
    function GetAsExpressionString: string; override;
  public
    constructor Create(AValue: ValReal);
    function Clone: TEvaluationValue; override;
    function TryGetAsFloat(out Float: ValReal; out ErrMessage: string): Boolean; override;
  end;

  { TEvaluationIdentifier }

  TEvaluationIdentifier = class;
  TCallParamsFuncProc = function(const Sender: TEvaluationIdentifier; const ParamValues: array of TEvaluationValue; out ErrMessage: string): TEvaluationValue of object;
  TGetValueProc = function(const Sender: TEvaluationIdentifier; out ErrMessage: string): TEvaluationValue of object;

  TEvaluationIdentifier = class
  private
    FName: string;
    FValue: TEvaluationValue;
    FOnCallParamsFunc: TCallParamsFuncProc;
    FOnGetValue: TGetValueProc;
    procedure SetValue(AValue: TEvaluationValue);
  public
    constructor Create(AName: string);
    destructor Destroy; override;
    property Name: string read FName write FName;
    property Value: TEvaluationValue read FValue write SetValue;
    property OnCallParamsFunc: TCallParamsFuncProc read FOnCallParamsFunc write FOnCallParamsFunc;
    property OnGetValue: TGetValueProc read FOnGetValue write FOnGetValue;
  end;

  TGenEvaluationIdentifierList = specialize TObjectList<TEvaluationIdentifier>;

  { TEvaluationIdentifierList }

  TEvaluationIdentifierList = class(TGenEvaluationIdentifierList)
  public
    function FindByName(AName: string): TEvaluationIdentifier;
  end;

  { TExpressionEvaluator }

  TExpressionEvaluator = class
  private
    FEngine: TPasTreeContainer;
    FExpression: string;
    FPasExpression: TPasExpr;
    FError: string;
  protected
    FIdentifierList: TEvaluationIdentifierList;

    function ExtractIdentifiers(AnExpression: TPasExpr): Boolean;

    function EvaluateExpr(AnExpression: TPasExpr): TEvaluationValue;
    function TryEvaluateBinaryArithmeticOperator(const ALeftValue, ARightValue: TEvaluationValue; const OpCode: TExprOpCode): TEvaluationValue;
    function TryEvaluateBinaryLogicalOperator(const ALeftValue, ARightValue: TEvaluationValue; const OpCode: TExprOpCode): TEvaluationValue;
    function TryEvaluateUnaryOperator(const AnOperand: TEvaluationValue; const OpCode: TExprOpCode): TEvaluationValue;

    procedure SetError(AnErrorMsg: string); overload;
    procedure SetError(Const AnErrorMsg: string; const Args: array of const); overload;
  public
    constructor Create(AnExpression: string);
    destructor Destroy; override;
    function Parse: Boolean;
    function Evaluate: TEvaluationValue;
    procedure AddMathFunctions;
    property IdentifierList: TEvaluationIdentifierList read FIdentifierList;
    property Error: string read FError;
  end;


  { TMonadicRealFunction }

  TRealFunc = function(V: ValReal):ValReal;

  TMonadicRealFunction = class(TEvaluationIdentifier)
    RealFunc: TRealFunc;
    function GetValue(const Sender: TEvaluationIdentifier; const ParamValues: array of TEvaluationValue; out ErrMessage: string): TEvaluationValue;
    constructor Create(AName: string; F: TRealFunc);
  end;

  TIntegerFunc = function(V: Int64):Int64;

  { TMonadicRealOrIntegerFunction }

  TMonadicRealOrIntegerFunction = class(TMonadicRealFunction)
    IntegerFunc: TIntegerFunc;
    function GetValue(const Sender: TEvaluationIdentifier; const ParamValues: array of TEvaluationValue; out ErrMessage: string): TEvaluationValue;
    constructor Create(AName: string; F: TRealFunc; FI: TIntegerFunc);
  end;

  { TMonadicRealFunctionI }

  TRealFuncI = function(V: ValReal):Int64;

  TMonadicRealFunctionI = class(TEvaluationIdentifier)
    RealFuncI: TRealFuncI;
    function GetValue(const Sender: TEvaluationIdentifier; const ParamValues: array of TEvaluationValue; out ErrMessage: string): TEvaluationValue;
    constructor Create(AName: string; F: TRealFuncI);
  end;

  TRealFunc2 = function(A, B: ValReal):ValReal;

  { TDiadicRealFunction }

  TDiadicRealFunction = class(TEvaluationIdentifier)
    RealFunc2: TRealFunc2;
    function GetValue(const Sender: TEvaluationIdentifier; const ParamValues: array of TEvaluationValue; out ErrMessage: string): TEvaluationValue;
    constructor Create(AName: string; F: TRealFunc2);
  end;

implementation

type
  { TSimpleEngine }

  { We have to override abstract TPasTreeContainer methods.
    See utils/fpdoc/dglobals.pp for an implementation of TFPDocEngine,
    a "real" engine. }

  TSimpleEngine = class(TPasTreeContainer)
  public
    function CreateElement(
      AClass: TPTreeElement;
      const AName: String;
      AParent: TPasElement;
      AVisibility: TPasMemberVisibility;
      const ASourceFilename: String;
      ASourceLinenumber: Integer): TPasElement; override;
    function FindElement(const AName: String): TPasElement; override;
  end;

{ TDiadicRealFunction }

function TDiadicRealFunction.GetValue(const Sender: TEvaluationIdentifier;
  const ParamValues: array of TEvaluationValue; out ErrMessage: string
  ): TEvaluationValue;
var
  A, B: ValReal;
begin
  Result := nil;
  if Length(ParamValues) <> 2 then
    ErrMessage := '2 parameters expected, but '+IntToStr(Length(ParamValues))+' given for '+Name
  else if not ParamValues[0].TryGetAsFloat(A, ErrMessage) then
    ErrMessage := 'Invalid argument 1 for '+Name+':'+ErrMessage
  else if not ParamValues[1].TryGetAsFloat(B, ErrMessage) then
    ErrMessage := 'Invalid argument 2 for '+Name+':'+ErrMessage
  else
    Result := TFloatEvaluationValue.Create(RealFunc2(A,B));
end;

constructor TDiadicRealFunction.Create(AName: string; F: TRealFunc2);
begin
  inherited Create(AName);
  RealFunc2 := F;
  OnCallParamsFunc := @GetValue;
end;

{ TMonadicRealFunction }

function TMonadicRealFunction.GetValue(const Sender: TEvaluationIdentifier;
  const ParamValues: array of TEvaluationValue; out ErrMessage: string
  ): TEvaluationValue;
var
  A: ValReal;
begin
  Result := nil;
  if Length(ParamValues) <> 1 then
    ErrMessage := 'Only 1 parameter expected, but '+IntToStr(Length(ParamValues))+' given for '+Name
  else if not ParamValues[0].TryGetAsFloat(A, ErrMessage) then
    ErrMessage := 'Invalid argument for '+Name+':'+ErrMessage
  else
    Result := TFloatEvaluationValue.Create(RealFunc(A));
end;

constructor TMonadicRealFunction.Create(AName: string; F: TRealFunc);
begin
  inherited Create(AName);
  RealFunc := F;
  OnCallParamsFunc := @GetValue;
end;

{ TMonadicRealOrIntegerFunction }

function TMonadicRealOrIntegerFunction.GetValue(
  const Sender: TEvaluationIdentifier;
  const ParamValues: array of TEvaluationValue; out ErrMessage: string
  ): TEvaluationValue;
var
  A: ValReal;
  I: Int64;
begin
  Result := nil;
  if Length(ParamValues) <> 1 then
    ErrMessage := 'Only 1 parameter expected, but '+IntToStr(Length(ParamValues))+' given for '+Name
  else if ParamValues[0].TryGetAsInteger(I, ErrMessage) then
    Result := TIntegerEvaluationValue.Create(IntegerFunc(I))
  else if not ParamValues[0].TryGetAsFloat(A, ErrMessage) then
    ErrMessage := 'Invalid argument for '+Name+':'+ErrMessage
  else
    Result := TFloatEvaluationValue.Create(RealFunc(A));
end;

constructor TMonadicRealOrIntegerFunction.Create(AName: string; F: TRealFunc; FI: TIntegerFunc);
begin
  inherited Create(AName, F);
  IntegerFunc := FI;
  OnCallParamsFunc := @GetValue;
end;

{ TMonadicRealFunctionI }

function TMonadicRealFunctionI.GetValue(const Sender: TEvaluationIdentifier;
  const ParamValues: array of TEvaluationValue; out ErrMessage: string
  ): TEvaluationValue;
var
  A: ValReal;
begin
  Result := nil;
  if Length(ParamValues) <> 1 then
    ErrMessage := 'Only 1 parameter expected, but '+IntToStr(Length(ParamValues))+' given for '+Name
  else if not ParamValues[0].TryGetAsFloat(A, ErrMessage) then
    ErrMessage := 'Invalid argument for '+Name+':'+ErrMessage
  else
    Result := TIntegerEvaluationValue.Create(RealFuncI(A));
end;

constructor TMonadicRealFunctionI.Create(AName: string; F: TRealFuncI);
begin
  inherited Create(AName);
  RealFuncI := F;
  OnCallParamsFunc := @GetValue;
end;

{ TFloatEvaluationValue }

constructor TFloatEvaluationValue.Create(AValue: ValReal);
begin
  FKind := evkFloat;
  FValue := AValue;
end;

function TFloatEvaluationValue.Clone: TEvaluationValue;
begin
  Result := TFloatEvaluationValue.Create(FValue);
end;

function TFloatEvaluationValue.TryGetAsFloat(out Float: ValReal; out ErrMessage: string): Boolean;
begin
  Result := True;
  Float := FValue;
  ErrMessage := '';
end;

function TFloatEvaluationValue.GetAsExpressionString: string;
begin
  Result := FloatToStr(FValue);
end;

{ TCardinalEvaluationValue }

constructor TCardinalEvaluationValue.Create(AValue: QWord);
begin
  FKind := evkCardinal;
end;

function TCardinalEvaluationValue.Clone: TEvaluationValue;
begin
  Result := TCardinalEvaluationValue.Create(FValue);
end;

function TCardinalEvaluationValue.TryGetAsCardinal(out Cardinal: QWord; out ErrMessage: string): Boolean;
begin
  Result := True;
  ErrMessage := '';
  Cardinal := FValue;
end;

function TCardinalEvaluationValue.TryGetAsInteger(out Int: Int64; out ErrMessage: string): Boolean;
begin
  if FValue < High(Int64) then
    begin
    Result := True;
    ErrMessage := '';
    Int := FValue;
    end
  else
    begin
    Result := False;
    ErrMessage := 'Range check error';
    Int := 0;
    end;
end;

function TCardinalEvaluationValue.TryGetAsFloat(out Float: ValReal; out ErrMessage: string): Boolean;
begin
  Result := True;
  ErrMessage := '';
  Float := FValue;
end;

function TCardinalEvaluationValue.GetAsExpressionString: string;
begin
  Result := IntToStr(FValue);
end;

{ TBooleanEvaluationValue }

constructor TBooleanEvaluationValue.Create(AValue: Boolean);
begin
  FKind := evkBoolean;
  FValue := AValue;
end;

function TBooleanEvaluationValue.Clone: TEvaluationValue;
begin
  Result := TBooleanEvaluationValue.Create(FValue);
end;

function TBooleanEvaluationValue.TryGetAsBoolean(out Bool: Boolean; out ErrMessage: string): Boolean;
begin
  ErrMessage := '';
  Result := True;
  Bool := FValue;
end;

function TBooleanEvaluationValue.GetAsExpressionString: string;
begin
  if FValue then
    Result := 'True'
  else
    Result := 'False'
end;

{ TStringEvaluationValue }

constructor TStringEvaluationValue.Create(AValue: string);
begin
  FKind := evkString;
  FValue := AValue;
end;

function TStringEvaluationValue.Clone: TEvaluationValue;
begin
  Result := TStringEvaluationValue.Create(FValue);
end;

function TStringEvaluationValue.TryGetAsString(out Str: string; out ErrMessage: string): Boolean;
begin
  ErrMessage := '';
  Result := True;
  Str := FValue;
end;

function TStringEvaluationValue.GetAsExpressionString: string;
begin
  Result := AnsiQuotedStr(FValue, '''');
end;

{ TEvaluationIdentifierList }

function TEvaluationIdentifierList.FindByName(AName: string): TEvaluationIdentifier;
var
  i: Integer;
begin
  Result := nil;
  for i := 0 to Count -1 do
    if SameText(Items[i].Name, AName) then
      begin
      Result := Items[i];
      Exit;
      end;
end;

{ TEvaluationIdentifier }

constructor TEvaluationIdentifier.Create(AName: string);
begin
  FName := AName;
end;

destructor TEvaluationIdentifier.Destroy;
begin
  FValue.Free;
  inherited Destroy;
end;

procedure TEvaluationIdentifier.SetValue(AValue: TEvaluationValue);
begin
  if Assigned(FValue) then
    raise Exception.Create('Value already assigned');
  FValue := AValue;
end;

{ TIntegerEvaluationValue }

constructor TIntegerEvaluationValue.Create(AValue: Int64);
begin
  FKind := evkInteger;
  FValue := AValue;
end;

function TIntegerEvaluationValue.TryGetAsInteger(out Int: Int64; out ErrMessage: string): Boolean;
begin
  ErrMessage := '';
  Int := FValue;
  Result := True;
end;

function TIntegerEvaluationValue.TryGetAsFloat(out Float: ValReal; out ErrMessage: string): Boolean;
begin
  ErrMessage := '';
  Float := FValue;
  Result := True;
end;

function TIntegerEvaluationValue.Clone: TEvaluationValue;
begin
  Result := TIntegerEvaluationValue.Create(FValue);
end;

function TIntegerEvaluationValue.GetAsExpressionString: string;
begin
  Result := IntToStr(FValue);
end;

{ TEvaluationValue }

function TEvaluationValue.TryGetAsString(out Str: string; out ErrMessage: string): Boolean;
begin
  Str := '';
  Result := False;
  ErrMessage := 'Unable to convert value to string';
end;

function TEvaluationValue.TryGetAsInteger(out Int: Int64; out ErrMessage: string): Boolean;
begin
  Int := 0;
  Result := False;
  ErrMessage := 'Unable to convert value to integer';
end;

function TEvaluationValue.Clone: TEvaluationValue;
begin
  Result := TEvaluationValue.Create;
  Result.FKind := FKind;
end;

function TEvaluationValue.TryGetAsBoolean(out Bool: Boolean; out ErrMessage: string): Boolean;
begin
  Bool := False;
  Result := False;
  ErrMessage := 'Unable to convert value to boolean';
end;

function TEvaluationValue.TryGetAsCardinal(out Cardinal: QWord; out ErrMessage: string): Boolean;
begin
  Cardinal := 0;
  Result := False;
  ErrMessage := 'Unable to convert value to cardinal';
end;

function TEvaluationValue.TryGetAsFloat(out Float: ValReal; out ErrMessage: string): Boolean;
begin
  Float := 0;
  Result := False;
  ErrMessage := 'Unable to convert value to float';
end;

function TEvaluationValue.GetAsExpressionString: string;
begin
  Result := '';
end;

function TSimpleEngine.CreateElement(AClass: TPTreeElement; const AName: string;
  AParent: TPasElement; AVisibility: TPasMemberVisibility; const ASourceFilename: string;
  ASourceLinenumber: Integer): TPasElement;
begin
  Result := AClass.Create(AName, AParent);
  Result.Visibility := AVisibility;
  Result.SourceFilename := ASourceFilename;
  Result.SourceLinenumber := ASourceLinenumber;
end;

function TSimpleEngine.FindElement(const AName: string): TPasElement;
begin
  { dummy implementation, see TFPDocEngine.FindElement for a real example }
  Result := nil;
end;

{ TExpressionEvaluator }

constructor TExpressionEvaluator.Create(AnExpression: string);
begin
  FEngine := TSimpleEngine.Create;
  FExpression := AnExpression;
  FIdentifierList := TEvaluationIdentifierList.Create;
end;

destructor TExpressionEvaluator.Destroy;
begin
  FIdentifierList.Free;
  FPasExpression.Release;
  FEngine.Free;
  inherited Destroy;
end;

function TExpressionEvaluator.Evaluate: TEvaluationValue;
begin
  Result := EvaluateExpr(FPasExpression);
end;

{ Required as can't take address of internal routines }

function _Sin(a:ValReal):ValReal;
begin
  Result := Sin(a);
end;

function _Cos(a:ValReal):ValReal;
begin
  Result := Cos(a);
end;

function _Atan(a:ValReal):ValReal;
begin
  Result := ArcTan(a);
end;

function _Exp(a:ValReal):ValReal;
begin
  Result := Exp(a);
end;

function _Ln(a:ValReal):ValReal;
begin
  Result := Ln(a);
end;

function _Sqrt(a:ValReal):ValReal;
begin
  Result := Sqrt(a);
end;

function _Sqr(a:ValReal):ValReal;
begin
  Result := Sqr(a);
end;

function _Ceil(a:Double):Double;
begin
  Result := Ceil(a);
end;

function _Floor(a:Double):Double;
begin
  Result := Floor(a);
end;

function _Round(a:ValReal):Int64;
begin
  Result := Round(a);
end;

function _Frac(a:ValReal):ValReal;
begin
  Result := Frac(a);
end;

function _Int(a:ValReal):ValReal;
begin
  Result := Int(a);
end;

function _Trunc(a:ValReal):int64;
begin
  Result := Trunc(a);
end;

function _AbsI(a:Int64):Int64;
begin
  Result := Abs(a);
end;

function _Abs(a:ValReal):ValReal;
begin
  Result := Abs(a);
end;

procedure TExpressionEvaluator.AddMathFunctions;
begin
  FIdentifierList.Add(TMonadicRealOrIntegerFunction.Create('abs',@_Abs,@_AbsI));

  FIdentifierList.Add(TMonadicRealFunction.Create('sqrt',@_Sqrt));
  FIdentifierList.Add(TMonadicRealFunction.Create('sqr',@_Sqr));

  FIdentifierList.Add(TMonadicRealFunction.Create('frac',@_Frac));
  FIdentifierList.Add(TMonadicRealFunction.Create('int',@_Int));

  FIdentifierList.Add(TMonadicRealFunctionI.Create('ceil',@Ceil64));
  FIdentifierList.Add(TMonadicRealFunctionI.Create('floor',@Floor64));
  FIdentifierList.Add(TMonadicRealFunctionI.Create('round',@_Round));
  FIdentifierList.Add(TMonadicRealFunctionI.Create('trunc',@_Trunc));

  FIdentifierList.Add(TMonadicRealFunction.Create('ln',@_Ln));
  FIdentifierList.Add(TMonadicRealFunction.Create('exp',@_Exp));
  FIdentifierList.Add(TMonadicRealFunction.Create('log10',@Log10));

  FIdentifierList.Add(TMonadicRealFunction.Create('sin',@_Sin));
  FIdentifierList.Add(TMonadicRealFunction.Create('cos',@_Cos));
  FIdentifierList.Add(TMonadicRealFunction.Create('tan',@Tan));

  FIdentifierList.Add(TMonadicRealFunction.Create('arcsin',@ArcSin));
  FIdentifierList.Add(TMonadicRealFunction.Create('arccos',@ArcCos));
  FIdentifierList.Add(TMonadicRealFunction.Create('arctan',@_ATan));

  FIdentifierList.Add(TMonadicRealFunction.Create('sinh',@SinH));
  FIdentifierList.Add(TMonadicRealFunction.Create('cosh',@CosH));
  FIdentifierList.Add(TMonadicRealFunction.Create('tanh',@TanH));

  FIdentifierList.Add(TMonadicRealFunction.Create('arcsinh',@ArcSinH));
  FIdentifierList.Add(TMonadicRealFunction.Create('arccosh',@ArcCosH));
  FIdentifierList.Add(TMonadicRealFunction.Create('arctanh',@ArcTanH));

  FIdentifierList.Add(TDiadicRealFunction.Create('power',@Power));
  FIdentifierList.Add(TDiadicRealFunction.Create('ArcTan2',@ArcTan2));
end;

function TExpressionEvaluator.Parse: Boolean;
var
  FileResolver: TStreamResolver;
  Scanner: TPascalScanner;
  Parser: TPasParser;
begin
  Result := False;
  if Assigned(FPasExpression) then
    raise Exception.Create('TcnocExpressionEvaluator.Parse can not be called twice');

  FileResolver := nil;
  Scanner := nil;
  Parser := nil;
  try
    FileResolver := TStreamResolver.Create;
    FileResolver.OwnsStreams:=true;

    FileResolver.AddStream('Expression', TStringStream.Create(FExpression));

    Scanner := TPascalScanner.Create(FileResolver);
    Scanner.LogEvents:=FEngine.ScannerLogEvents;
    Scanner.OnLog:=FEngine.Onlog;

    Parser := TPasParser.Create(Scanner, FileResolver, FEngine);
    Parser.LogEvents:=FEngine.ParserLogEvents;
    Parser.OnLog:=FEngine.Onlog;

    Scanner.OpenFile('Expression');

    parser.NextToken;
    parser.ParseAdhocExpression(FPasExpression);

    Result := ExtractIdentifiers(FPasExpression);
  finally
    Parser.Free;
    Scanner.Free;
    FileResolver.Free;
  end;
end;


function TExpressionEvaluator.EvaluateExpr(AnExpression: TPasExpr): TEvaluationValue;
var
  IdentName, Number: string;
  LeftExpr, RightExpr: TEvaluationValue;
  Identifier: TEvaluationIdentifier;
  ParamsExpr: TParamsExpr;
  ParamValues: array of TEvaluationValue;
  i: Integer;
  i64: Int64;
begin
  Result := nil;
  if not assigned(AnExpression) then
    begin
    SetError('Incomplete expression');
    Exit;
    end;
  case AnExpression.Kind of
    pekBoolConst:
      Result := TBooleanEvaluationValue.Create((AnExpression as TBoolConstExpr).Value);
    pekNumber:
      begin
        Number := (AnExpression as TPrimitiveExpr).Value;
        if TryStrToInt64(Number, i64) then
          Result := TIntegerEvaluationValue.Create(i64)
        else
          Result := TFloatEvaluationValue.Create(StrToFloat(Number));
      end;
    pekString:
      Result := TStringEvaluationValue.Create(AnsiDequotedStr((AnExpression as TPrimitiveExpr).Value, ''''));
    pekBinary:
      begin
      LeftExpr := EvaluateExpr(TBinaryExpr(AnExpression).left);
      try
        if not Assigned(LeftExpr) then
          Exit;
        RightExpr := EvaluateExpr(TBinaryExpr(AnExpression).right);
        try
          if not Assigned(RightExpr) then
            Exit;
          case AnExpression.OpCode of
            eopAdd,
            eopSubtract,
            eopMultiply,
            eopDivide,
            eopDiv,
            eopMod,
            eopPower,
            eopShl, eopShr, eopAnd, eopOr, eopXor:
              Result := TryEvaluateBinaryArithmeticOperator(LeftExpr, RightExpr, AnExpression.OpCode);
            eopEqual,
            eopNotEqual,
            eopLessThan,
            eopGreaterThan,
            eopLessthanEqual,
            eopGreaterThanEqual:
              Result := TryEvaluateBinaryLogicalOperator(LeftExpr, RightExpr, AnExpression.OpCode);
          else
            SetError('Unsupported opcode: '+OpcodeStrings[AnExpression.OpCode]);
          end;
        finally
          RightExpr.Free;
        end;
      finally
        LeftExpr.Free;
      end;
      end;
      //
      //  //eopEqual:    result := EvaluateExpr(ASession, TBinaryExpr(AnExpression).left) = EvaluateExpr(ASession, TBinaryExpr(AnExpression).right);
      //  //eopNotEqual: result := EvaluateExpr(ASession, TBinaryExpr(AnExpression).left) <> EvaluateExpr(ASession, TBinaryExpr(AnExpression).right);
      //  //eopOr:       result := EvaluateExpr(ASession, TBinaryExpr(AnExpression).left) or EvaluateExpr(ASession, TBinaryExpr(AnExpression).right);
      //  //eopAnd:      result := EvaluateExpr(ASession, TBinaryExpr(AnExpression).left) and EvaluateExpr(ASession, TBinaryExpr(AnExpression).right);
      //  //eopGreaterThan: result := EvaluateExpr(ASession, TBinaryExpr(AnExpression).left) > EvaluateExpr(ASession, TBinaryExpr(AnExpression).right);
      //  //eopLessThan: result := EvaluateExpr(ASession, TBinaryExpr(AnExpression).left) < EvaluateExpr(ASession, TBinaryExpr(AnExpression).right);
      //else
      //  raise exception.CreateFmt('Unsupported opcode (%s) in expression', [OpcodeStrings[AnExpression.OpCode]]);
      //end;
      //end;
    pekUnary:
      begin
      LeftExpr := EvaluateExpr(TUnaryExpr(AnExpression).Operand);
      try
        if not Assigned(LeftExpr) then
          Exit;
        Result := TryEvaluateUnaryOperator(LeftExpr, AnExpression.OpCode);
      finally
        LeftExpr.Free;
      end;
      end;
    pekIdent:
      begin
      IdentName:=(AnExpression as TPrimitiveExpr).Value;
      Identifier := FIdentifierList.FindByName(IdentName);
      if not Assigned(Identifier) then
        // May never happen
        raise Exception.Create('Evaluation problem, identifier not found.');
      if Assigned(Identifier.OnGetValue) then
        Result := Identifier.OnGetValue(Identifier, FError)
      else
        begin
        LeftExpr := Identifier.Value;
        if not Assigned(LeftExpr) then
          SetError('No value provided for identifier [%s]', [IdentName])
        else
          Result := LeftExpr.Clone;
        end;
      end;
    pekFuncParams:
      begin
      ParamsExpr := AnExpression as TParamsExpr;

      IdentName:=(ParamsExpr.Value as TPrimitiveExpr).Value;
      Identifier := FIdentifierList.FindByName(IdentName);
      if not Assigned(Identifier) then
        // May never happen
        raise Exception.Create('Evaluation problem, function not found.');
      if not Assigned(Identifier.OnCallParamsFunc) then
        begin
        SetError('Calling [%s] is not supported', [IdentName]);
        Exit;
        end;

      SetLength(ParamValues, Length(ParamsExpr.Params));
      try
        for i := 0 to High(ParamsExpr.Params) do
          begin
          ParamValues[i] := EvaluateExpr(ParamsExpr.Params[i]);
          if not Assigned(ParamValues[i]) then
            Exit;
          end;
        Result := Identifier.OnCallParamsFunc(Identifier, ParamValues, FError);
      finally
        for i := 0 to High(ParamsExpr.Params) do
          ParamValues[i].Free;
      end;
      end;
  else
    begin
    SetError('Unsupported expression (%s) in expression', [AnExpression.ClassName]);
    end;
  end;
end;

procedure TExpressionEvaluator.SetError(AnErrorMsg: string);
begin
  FError := AnErrorMsg;
end;

procedure TExpressionEvaluator.SetError(const AnErrorMsg: string; const Args: array of const);
begin
  SetError(Format(AnErrorMsg, Args));
end;

function TExpressionEvaluator.TryEvaluateBinaryArithmeticOperator(const ALeftValue, ARightValue: TEvaluationValue; const OpCode: TExprOpCode): TEvaluationValue;
type
  TTypeCastResultKind = (tcrInvalid, tcrUnsupported, tcrInteger, tcrString, tcrFloat, tcrBoolean);
const
  TypeCastArr: array[TEvaluationValueKind, TEvaluationValueKind] of TTypeCastResultKind =
    (           {evkNone        evkString      evkBoolean     evkInteger     evkCardinal    evkFloat }
{ evkNone     } (tcrUnsupported,tcrUnsupported,tcrUnsupported,tcrUnsupported,tcrUnsupported,tcrUnsupported),
{ evkString   } (tcrUnsupported,tcrString,     tcrUnsupported,tcrInvalid,    tcrInvalid,    tcrInvalid),
{ evkBoolean  } (tcrUnsupported,tcrUnsupported,tcrBoolean,    tcrUnsupported,tcrUnsupported,tcrUnsupported),
{ evkInteger  } (tcrUnsupported,tcrInvalid,    tcrUnsupported,tcrInteger,    tcrInteger,    tcrFloat),
{ evkCardinal } (tcrUnsupported,tcrInvalid,    tcrUnsupported,tcrInteger,    tcrInteger,    tcrUnsupported),
{ evkFloat    } (tcrUnsupported,tcrInvalid,    tcrUnsupported,tcrFloat,      tcrUnsupported,tcrFloat)
    );
var
  TypeCastResultKind: TTypeCastResultKind;
  LeftValueInt, RightValueInt: Int64;
  LeftValueFloat, RightValueFloat: ValReal;
  LeftValueBool, RightValueBool: Boolean;
  RightValueString, LeftValueString: string;
  LeftKind, RightKind: TEvaluationValueKind;
  ErrMsg: string;
begin
  Result := nil;
  LeftKind := ALeftValue.Kind;
  RightKind := ARightValue.Kind;
  TypeCastResultKind := TypeCastArr[LeftKind, RightKind];
  case TypeCastResultKind of
    tcrInvalid:
      begin
      SetError('Incompatible types');
      Exit;
      end;
    tcrUnsupported:
      begin
      SetError('not-implemented typecase');
      Exit;
      end;
    tcrInteger:
      begin
      if not ALeftValue.TryGetAsInteger(LeftValueInt, ErrMsg) or
         not ARightValue.TryGetAsInteger(RightValueInt, ErrMsg) then
        begin
        SetError(ErrMsg);
        Exit;
        end;
      case OpCode of
        eopAdd:      LeftValueInt := LeftValueInt + RightValueInt;
        eopSubtract: LeftValueInt := LeftValueInt - RightValueInt;
        eopMultiply: LeftValueInt := LeftValueInt * RightValueInt;
        eopDiv:      LeftValueInt := LeftValueInt div RightValueInt;
        eopMod:      LeftValueInt := LeftValueInt mod RightValueInt;
        eopShl:      LeftValueInt := LeftValueInt shl RightValueInt;
        eopShr:      LeftValueInt := LeftValueInt shr RightValueInt;
        eopAnd:      LeftValueInt := LeftValueInt and RightValueInt;
        eopOr:       LeftValueInt := LeftValueInt or  RightValueInt;
        eopXor:      LeftValueInt := LeftValueInt xor  RightValueInt;
        eopDivide:   begin
                       Result := TFloatEvaluationValue.Create(LeftValueInt/RightValueInt);
                       Exit;
                     end
      else
        begin
        SetError('Unsupported integer opcode: '+OpcodeStrings[OpCode]);
        Exit;
        end;
      end;
      Result := TIntegerEvaluationValue.Create(LeftValueInt);
      end;
    tcrFloat:
      begin
      if not ALeftValue.TryGetAsFloat(LeftValueFloat, ErrMsg) or
         not ARightValue.TryGetAsFloat(RightValueFloat, ErrMsg) then
        begin
        SetError(ErrMsg);
        Exit;
        end;
      case OpCode of
        eopAdd:      LeftValueFloat := LeftValueFloat + RightValueFloat;
        eopSubtract: LeftValueFloat := LeftValueFloat - RightValueFloat;
        eopMultiply: LeftValueFloat := LeftValueFloat * RightValueFloat;
        eopDivide:   LeftValueFloat := LeftValueFloat / RightValueFloat;
      else
        begin
        SetError('Unsupported floating point opcode: '+OpcodeStrings[OpCode]);
        Exit;
        end;
      end;
      Result := TFloatEvaluationValue.Create(LeftValueFloat);
      end;
    tcrString:
      begin
      if not ALeftValue.TryGetAsString(LeftValueString, ErrMsg) or
         not ARightValue.TryGetAsString(RightValueString, ErrMsg) then
        begin
        SetError(ErrMsg);
        Exit;
        end;
      case OpCode of
        eopAdd:      LeftValueString := LeftValueString + RightValueString;
      else
        begin
        SetError('Unsupported string opcode: '+OpcodeStrings[OpCode]);
        Exit;
        end;
      end;
      Result := TStringEvaluationValue.Create(LeftValueString);
      end;
    tcrBoolean:
      begin
      if not ALeftValue.TryGetAsBoolean(LeftValueBool, ErrMsg) or
         not ARightValue.TryGetAsBoolean(RightValueBool, ErrMsg) then
        begin
        SetError(ErrMsg);
        Exit;
        end;
      case OpCode of
        eopAnd:      LeftValueBool := LeftValueBool and RightValueBool;
        eopOr:       LeftValueBool := LeftValueBool or  RightValueBool;
        eopXor:      LeftValueBool := LeftValueBool xor  RightValueBool;
      else
        begin
        SetError('Unsupported boolean opcode: '+OpcodeStrings[OpCode]);
        Exit;
        end;
      end;
      Result := TBooleanEvaluationValue.Create(LeftValueBool);
      end;
  else
    // Should be dead code
    raise Exception.Create('Failed to evaluate binary operator');
  end;
end;

function TExpressionEvaluator.ExtractIdentifiers(AnExpression: TPasExpr): Boolean;
var
  AParamsExpr: TParamsExpr;
  i: Integer;
  IdentName: string;
begin
  Result := False;
  if not Assigned(AnExpression) then
    begin
    SetError('Failed to extract identifier: expression is missing');
    Exit;
    end;

  case AnExpression.Kind of
    pekBoolConst,
    pekNumber,
    pekString:
      begin
      // No identifier
      Result := True;
      end;
    pekBinary:
      begin
      Result := ExtractIdentifiers(TBinaryExpr(AnExpression).left) and ExtractIdentifiers(TBinaryExpr(AnExpression).right);
      end;
    pekIdent:
      begin
      IdentName := (AnExpression as TPrimitiveExpr).Value;
      if not assigned(FIdentifierList.FindByName(IdentName)) then
        FIdentifierList.Add(TEvaluationIdentifier.Create(IdentName));
      Result := True;
      end;
    pekUnary:
      begin
      Result := ExtractIdentifiers(TUnaryExpr(AnExpression).Operand);
      end;
    pekFuncParams:
      begin
      AParamsExpr := AnExpression as TParamsExpr;
      for i := 0 to High(AParamsExpr.Params) do
        begin
        Result := ExtractIdentifiers(AParamsExpr.Params[i]);
        if not Result then
          Exit;
        end;
      if Assigned(AParamsExpr.Value) then
        begin
        Result := ExtractIdentifiers(AParamsExpr.Value);
        if not Result then
          Exit;
        end;
      end;
  else
    SetError(Format('Unsupported expression (%s) in expression', [AnExpression.ClassName]));
  end;
end;

function TExpressionEvaluator.TryEvaluateUnaryOperator(const AnOperand: TEvaluationValue; const OpCode: TExprOpCode): TEvaluationValue;
var
  Err: string;
  Int: Int64;
  Bool: Boolean;
  Cardinal: QWord;
  Real: ValReal;
begin
  Result := nil;
  case OpCode of
    eopSubtract:
      begin
      case AnOperand.Kind of
        evkInteger:
          begin
          if AnOperand.TryGetAsInteger(Int, Err) then
            Result := TIntegerEvaluationValue.Create(-Int)
          else
            SetError(Err);
          end;
        evkFloat:
          begin
          if AnOperand.TryGetAsFloat(Real, Err) then
            Result := TFloatEvaluationValue.Create(-Real)
          else
            SetError(Err);
          end
      else
        SetError('Incompatible types');
      end;
      end;
    eopNot:
      begin
      case AnOperand.Kind of
        evkBoolean:
          begin
          if AnOperand.TryGetAsBoolean(Bool, Err) then
            Result := TBooleanEvaluationValue.Create(not Bool)
          else
            SetError(Err);
          end;
        evkInteger:
          begin
          if AnOperand.TryGetAsInteger(Int, Err) then
            Result := TIntegerEvaluationValue.Create(not Int)
          else
            SetError(Err);
          end;
        evkCardinal:
          begin
          if AnOperand.TryGetAsCardinal(Cardinal, Err) then
            Result := TCardinalEvaluationValue.Create(not Cardinal)
          else
            SetError(Err);
          end
      else
        SetError('Incompatible types');
      end;
      end;
  else
    SetError('Unsupported unary opcode');
  end;
end;

function TExpressionEvaluator.TryEvaluateBinaryLogicalOperator(const ALeftValue, ARightValue: TEvaluationValue; const OpCode: TExprOpCode): TEvaluationValue;
var
  LeftValueInt, RightValueInt: Int64;
  RightValueString, LeftValueString: string;
  RightValueBool, LeftValueBool: boolean;
  RightValueFloat, LeftValueFloat: ValReal;
  RightValueCardinal, LeftValueCardinal: QWord;
  ErrMsg: string;
  ResultBool: Boolean;
  LeftKind, RightKind: TEvaluationValueKind;
const
  TypeCastArr: array[TEvaluationValueKind, TEvaluationValueKind] of TEvaluationValueKind =
    (           {evkNone evkString  evkBoolean evkInteger evkCardinal evkFloat }
{ evkNone     } (evkNone,evkNone,   evkNone,   evkNone,   evkNone,    evkNone),
{ evkString   } (evkNone,evkString, evkNone,   evkNone,   evkNone,    evkNone),
{ evkBoolean  } (evkNone,evkNone,   evkBoolean,evkNone,   evkNone,    evkNone),
{ evkInteger  } (evkNone,evkNone,   evkNone,   evkInteger,evkInteger, evkFloat),
{ evkCardinal } (evkNone,evkNone,   evkNone,   evkInteger,evkCardinal,evkFloat),
{ evkFloat    } (evkNone,evkNone,   evkNone,   evkFloat,  evkFloat,   evkFloat)
    );
begin
  Result := nil;
  LeftKind := ALeftValue.Kind;
  RightKind := ARightValue.Kind;
  case TypeCastArr[LeftKind, RightKind] of
    evkBoolean :
      begin
      if not ALeftValue.TryGetAsBoolean(LeftValueBool, ErrMsg) or
         not ARightValue.TryGetAsBoolean(RightValueBool, ErrMsg) then
        begin
        SetError(ErrMsg);
        Exit;
        end;
      case OpCode of
        eopEqual:            ResultBool := LeftValueBool = RightValueBool;
        eopNotEqual:         ResultBool := LeftValueBool <> RightValueBool;
        eopLessThan:         ResultBool := LeftValueBool < RightValueBool;
        eopLessthanEqual:    ResultBool := LeftValueBool <= RightValueBool;
        eopGreaterThan:      ResultBool := LeftValueBool > RightValueBool;
        eopGreaterThanEqual: ResultBool := LeftValueBool >= RightValueBool;
      else
        begin
        SetError('Unsupported logical boolean opcode');
        exit;
        end;
      end;
      end;
    evkInteger :
      begin
      if not ALeftValue.TryGetAsInteger(LeftValueInt, ErrMsg) or
         not ARightValue.TryGetAsInteger(RightValueInt, ErrMsg) then
        begin
        SetError(ErrMsg);
        Exit;
        end;
      case OpCode of
        eopEqual:            ResultBool := LeftValueInt = RightValueInt;
        eopNotEqual:         ResultBool := LeftValueInt <> RightValueInt;
        eopLessThan:         ResultBool := LeftValueInt < RightValueInt;
        eopLessthanEqual:    ResultBool := LeftValueInt <= RightValueInt;
        eopGreaterThan:      ResultBool := LeftValueInt > RightValueInt;
        eopGreaterThanEqual: ResultBool := LeftValueInt >= RightValueInt;
      else
        begin
        SetError('Unsupported logical boolean opcode');
        exit
        end;
      end;
      end;
    evkString :
      begin
      if not ALeftValue.TryGetAsString(LeftValueString, ErrMsg) or
         not ARightValue.TryGetAsString(RightValueString, ErrMsg) then
        begin
        SetError(ErrMsg);
        Exit;
        end;
      case OpCode of
        eopEqual:            ResultBool := LeftValueString = RightValueString;
        eopNotEqual:         ResultBool := LeftValueString <> RightValueString;
        eopLessThan:         ResultBool := LeftValueString < RightValueString;
        eopLessthanEqual:    ResultBool := LeftValueString <= RightValueString;
        eopGreaterThan:      ResultBool := LeftValueString > RightValueString;
        eopGreaterThanEqual: ResultBool := LeftValueString >= RightValueString;
      else
        begin
        SetError('Unsupported logical boolean opcode');
        exit
        end;
      end;
      end;
    evkFloat :
      begin
      if not ALeftValue.TryGetAsFloat(LeftValueFloat, ErrMsg) or
         not ARightValue.TryGetAsFloat(RightValueFloat, ErrMsg) then
        begin
        SetError(ErrMsg);
        Exit;
        end;
      case OpCode of
        eopEqual:            ResultBool := LeftValueFloat = RightValueFloat;
        eopNotEqual:         ResultBool := LeftValueFloat <> RightValueFloat;
        eopLessThan:         ResultBool := LeftValueFloat < RightValueFloat;
        eopLessthanEqual:    ResultBool := LeftValueFloat <= RightValueFloat;
        eopGreaterThan:      ResultBool := LeftValueFloat > RightValueFloat;
        eopGreaterThanEqual: ResultBool := LeftValueFloat >= RightValueFloat;
      else
        begin
        SetError('Unsupported logical boolean opcode');
        exit
        end;
      end;
      end;
    evkCardinal :
      begin
      if not ALeftValue.TryGetAsCardinal(LeftValueCardinal, ErrMsg) or
         not ARightValue.TryGetAsCardinal(RightValueCardinal, ErrMsg) then
        begin
        SetError(ErrMsg);
        Exit;
        end;
      case OpCode of
        eopEqual:            ResultBool := LeftValueCardinal = RightValueCardinal;
        eopNotEqual:         ResultBool := LeftValueCardinal <> RightValueCardinal;
        eopLessThan:         ResultBool := LeftValueCardinal < RightValueCardinal;
        eopLessthanEqual:    ResultBool := LeftValueCardinal <= RightValueCardinal;
        eopGreaterThan:      ResultBool := LeftValueCardinal > RightValueCardinal;
        eopGreaterThanEqual: ResultBool := LeftValueCardinal >= RightValueCardinal;
      else
        begin
        SetError('Unsupported logical boolean opcode');
        exit
        end;
      end;
      end;
  else
    begin
    SetError('Types can not be compared');
    exit
    end;
  end;
  Result := TBooleanEvaluationValue.Create(ResultBool);
end;

end.

