unit ExpressionTests;

{$mode objfpc}{$H+}

interface

uses
  Classes,
  SysUtils,
  fpcunit,
  testutils,
  testregistry,
  cnocExpressionEvaluator;

type

  { TTestExpressions }

  TTestExpressions = class(TTestCase)
  private
    FIdentifierList: TEvaluationIdentifierList;
    FWithMathFunctions: Boolean;
  protected
    procedure ProvideIdentifier(const AnIdentifier: TEvaluationIdentifier);
    procedure ProvideIdentifier(const AName: string; AnEvaluationValue: TEvaluationValue);
    procedure FillIdentifiers(AnEvaluator: TExpressionEvaluator);

    function TestExpression(const AnExpression: string; const AnExpectedKind: TEvaluationValueKind): TEvaluationValue;
    procedure IntTestEvaluationError(const AnExpression, AnExpectedErrorMessage: string);
    procedure IntTestExpressionIntegerResult(const AnExpression: string; AnExpectedResult: int64);
    procedure IntTestExpressionRealResult(const AnExpression: string; AnExpectedResult: ValReal);
    procedure IntTestExpressionStringResult(const AnExpression: string; AnExpectedResult: string);
    procedure IntTestExpressionBooleanResult(const AnExpression: string; AnExpectedResult: Boolean);
    procedure SetUp; override;
    procedure TearDown; override;
  published
    procedure TestBasicExpressions;
    procedure TestBasicExpressionsWithIdentifiers;
    procedure TestComparisionExpressions;
    procedure TestRealExpressions;
    procedure TestMathFunctions;
  end;

implementation

uses Math;

procedure TTestExpressions.TestBasicExpressions;
begin
  IntTestExpressionIntegerResult('1+1', 2);
  IntTestExpressionIntegerResult('123-51', 72);
  IntTestExpressionIntegerResult('3-5', -2);
  IntTestExpressionIntegerResult('3*5', 15);
  IntTestExpressionIntegerResult('9 div 2', 4);
  IntTestExpressionIntegerResult('3 * -5', -15);
  IntTestExpressionIntegerResult('(3+1)*5', 20);
  IntTestExpressionStringResult('''hello'' + '' '' + ''world''', 'hello world');
  IntTestExpressionStringResult('''It''''s'' + '' perfect''', 'It''s perfect');

  IntTestExpressionBooleanResult('true', True);
  IntTestExpressionBooleanResult('True', True);

  IntTestExpressionBooleanResult('false', False);
  IntTestExpressionBooleanResult('FALSE', False);

  IntTestExpressionBooleanResult('not true', False);
  IntTestExpressionBooleanResult('not false', True);

  IntTestExpressionIntegerResult('3 and 1', 1);
  IntTestExpressionIntegerResult('3 or 1',  3);
  IntTestExpressionIntegerResult('3 xor 1', 2);
  IntTestExpressionIntegerResult('3 shl 2', 12);
  IntTestExpressionIntegerResult('3 shr 1', 1);
  IntTestExpressionIntegerResult('not 1', not 1);
  IntTestExpressionIntegerResult('-5', -5);
  IntTestExpressionIntegerResult('-(3+1)*5', -20);
  IntTestExpressionIntegerResult('7 mod 3', 1);

end;

function TTestExpressions.TestExpression(const AnExpression: string; const AnExpectedKind: TEvaluationValueKind): TEvaluationValue;
var
  Evaluator: TExpressionEvaluator;
  Value: TEvaluationValue;
begin
  Evaluator := TExpressionEvaluator.Create(AnExpression);
  try
    if FWithMathFunctions then
      Evaluator.AddMathFunctions;
    Check(Evaluator.Parse, Format('Failed to parse expression [%s]: %s', [AnExpression, Evaluator.Error]));
    FillIdentifiers(Evaluator);
    Value := Evaluator.Evaluate;
    try
      Check(Assigned(Value), Format('Failed to evaluate expression [%s]: %s', [AnExpression, Evaluator.Error]));
      Check(AnExpectedKind = Value.Kind, Format('The resultkind is not what is expected for expression [%s]', [AnExpression]));
      Result := Value;
      Value := nil;
    finally
      Value.Free;
    end;
  finally
    Evaluator.Free;
  end;
end;

procedure TTestExpressions.IntTestExpressionIntegerResult(const AnExpression: string; AnExpectedResult: int64);
var
  Value: TEvaluationValue;
  ErrMessage: string;
  Int: Int64;
begin
  Value := TestExpression(AnExpression, evkInteger);
  try
    Check(Value.TryGetAsInteger(Int, ErrMessage), Format('Failed to convert evaluation-result of expression [%s] to integer: [%s]', [AnExpression, ErrMessage]));
    CheckEquals(AnExpectedResult, Int, Format('Result of expression [%s]', [AnExpression]));
  finally
    Value.Free;
  end;
end;

procedure TTestExpressions.IntTestExpressionRealResult(const AnExpression: string; AnExpectedResult: ValReal);
var
  Value: TEvaluationValue;
  ErrMessage: string;
  Real: ValReal;
begin
  Value := TestExpression(AnExpression, evkFloat);
  try
    Check(Value.TryGetAsFloat(Real, ErrMessage), Format('Failed to convert evaluation-result of expression [%s] to float: [%s]', [AnExpression, ErrMessage]));
    CheckEquals(AnExpectedResult, Real, Format('Result of expression [%s]', [AnExpression]));
  finally
    Value.Free;
  end;
end;

procedure TTestExpressions.TestBasicExpressionsWithIdentifiers;
begin
  IntTestEvaluationError('i', 'No value provided for identifier [i]');

  ProvideIdentifier('i', TIntegerEvaluationValue.Create(65));
  IntTestExpressionIntegerResult('i', 65);
  IntTestExpressionIntegerResult('i+5', 70);
  IntTestExpressionIntegerResult('i+I', 130);
end;

procedure TTestExpressions.IntTestEvaluationError(const AnExpression, AnExpectedErrorMessage: string);
var
  Evaluator: TExpressionEvaluator;
  Value: TEvaluationValue;
begin
  Evaluator := TExpressionEvaluator.Create(AnExpression);
  try
    Check(Evaluator.Parse, Format('Failed to parse expression [%s]: %s', [AnExpression, Evaluator.Error]));
    FillIdentifiers(Evaluator);
    Value := Evaluator.Evaluate;
    try
      Check(not Assigned(Value), Format('The evaluation provided a result, while an error was expected for expression [%s]', [AnExpression]));
      CheckEquals(AnExpectedErrorMessage, Evaluator.Error, Format('Expected error message for expression [%s]', [AnExpression]));
    finally
      Value.Free;
    end;
  finally
    Evaluator.Free;
  end;
end;

procedure TTestExpressions.ProvideIdentifier(const AName: string; AnEvaluationValue: TEvaluationValue);
var
  Identifier: TEvaluationIdentifier;
begin
  Identifier := TEvaluationIdentifier.Create(AName);
  Identifier.Value := AnEvaluationValue;
  ProvideIdentifier(Identifier);
end;

procedure TTestExpressions.ProvideIdentifier(const AnIdentifier: TEvaluationIdentifier);
begin
  FIdentifierList.Add(AnIdentifier);
end;

procedure TTestExpressions.SetUp;
begin
  inherited SetUp;
  FIdentifierList := TEvaluationIdentifierList.Create;
end;

procedure TTestExpressions.TearDown;
begin
  FIdentifierList.Free;
  inherited TearDown;
end;

procedure TTestExpressions.FillIdentifiers(AnEvaluator: TExpressionEvaluator);
var
  i: Integer;
  DestIdentifier: TEvaluationIdentifier;
  SourceIdentifier: TEvaluationIdentifier;
begin
  for i := 0 to AnEvaluator.IdentifierList.Count -1 do
    begin
    DestIdentifier := AnEvaluator.IdentifierList[i];
    SourceIdentifier := FIdentifierList.FindByName(DestIdentifier.Name);
    if Assigned(SourceIdentifier) then
      DestIdentifier.Value := SourceIdentifier.Value.Clone;
    end;
end;

procedure TTestExpressions.IntTestExpressionStringResult(const AnExpression: string; AnExpectedResult: string);
var
  Value: TEvaluationValue;
  ErrMessage: string;
  Str: string;
begin
  Value := TestExpression(AnExpression, evkString);
  try
    Check(Value.TryGetAsString(Str, ErrMessage), Format('Failed to convert evaluation-result of expression [%s] to string: [%s]', [AnExpression, ErrMessage]));
    CheckEquals(AnExpectedResult, Str, Format('Result of expression [%s]', [AnExpression]));
  finally
    Value.Free;
  end;
end;

procedure TTestExpressions.TestComparisionExpressions;
begin
  IntTestExpressionBooleanResult('1=1', True);
  IntTestExpressionBooleanResult('1=2', False);

  IntTestExpressionBooleanResult('1<>2', True);
  IntTestExpressionBooleanResult('1<>1', False);

  IntTestExpressionBooleanResult('1<2', True);
  IntTestExpressionBooleanResult('1<1', False);

  IntTestExpressionBooleanResult('1>2', False);
  IntTestExpressionBooleanResult('2>1', True);

  IntTestExpressionBooleanResult('''hoi''=''hoi''', True);
  IntTestExpressionBooleanResult('''hoi''=''dag''', False);

  IntTestExpressionBooleanResult('''hoi''<>''hoi''', False);
  IntTestExpressionBooleanResult('''hoi''<>''dag''', True);

  IntTestExpressionBooleanResult('''a''<''b''', True);
  IntTestExpressionBooleanResult('''B''>''A''', True);

  IntTestExpressionBooleanResult('1.0=1', True);
  IntTestExpressionBooleanResult('1.0=2', False);

  IntTestExpressionBooleanResult('1.0<>2', True);
  IntTestExpressionBooleanResult('1.0<>1', False);

  IntTestExpressionBooleanResult('1.0<2', True);
  IntTestExpressionBooleanResult('1.0<1', False);

  IntTestExpressionBooleanResult('1.0>2', False);
  IntTestExpressionBooleanResult('2.0>1', True);

  IntTestExpressionBooleanResult('1.0=1e0', True);
  IntTestExpressionBooleanResult('1.0=2e0', False);

  IntTestExpressionBooleanResult('1.0<>2.0', True);
  IntTestExpressionBooleanResult('1.0<>1.0', False);

  IntTestExpressionBooleanResult('1.0<2.0e0', True);
  IntTestExpressionBooleanResult('1.0<1.', False);

  IntTestExpressionBooleanResult('1.0>2.', False);
  IntTestExpressionBooleanResult('2.0>1.0e00', True);

  IntTestExpressionBooleanResult('(1 > 2) and (1 < 2)', False);
  IntTestExpressionBooleanResult('(1 < 2) and (1 > 2)', False);
  IntTestExpressionBooleanResult('(1 < 2) or (1 > 2)', True);
  IntTestExpressionBooleanResult('(1 < 2) xor (1 > 2)', True);

end;

procedure TTestExpressions.TestRealExpressions;
begin
  IntTestExpressionRealResult('1.1+2.0', 3.1);
  IntTestExpressionRealResult('1+2.0', 3.0);
  IntTestExpressionRealResult('1.1+2', 3.1);
  IntTestExpressionRealResult('1/3',   1/3);
  IntTestExpressionRealResult('2.0*5', 10);
  IntTestExpressionRealResult('-0.7', -0.7);
  IntTestExpressionRealResult('(1+2)/3', 1);
  IntTestExpressionRealResult('1-0.5', 0.5);
  IntTestExpressionRealResult('1*0.5', 0.5);
end;

procedure TTestExpressions.TestMathFunctions;
begin

  FWithMathFunctions := True;

  IntTestExpressionIntegerResult('Abs(-3)', 3);

  IntTestExpressionRealResult('Abs(-3.3)', Abs(3.3));

  IntTestExpressionRealResult('Sqrt(2)', Sqrt(2));
  IntTestExpressionRealResult('Sqr(2)', Sqr(2));

  IntTestExpressionRealResult('Sin(1)', Sin(1));
  IntTestExpressionRealResult('Cos(1)', Cos(1));
  IntTestExpressionRealResult('Tan(1)', Tan(1));

  IntTestExpressionRealResult('frac(5.3)', frac(5.3));
  IntTestExpressionRealResult('int(5.3)', Int(5.3));

  IntTestExpressionIntegerResult('ceil(5.3)',Ceil(5.3));
  IntTestExpressionIntegerResult('floor(5.3)', Floor(5.3));
  IntTestExpressionIntegerResult('round(5.3)', Round(5.3));
  IntTestExpressionIntegerResult('trunc(5.6)', Trunc(5.6));

  IntTestExpressionRealResult('ln(0.8)', Ln(0.8));
  IntTestExpressionRealResult('exp(0.8)', Exp(0.8));
  IntTestExpressionRealResult('log10(0.8)',Log10(0.8));

  IntTestExpressionRealResult('sin(0.8)',Sin(0.8));
  IntTestExpressionRealResult('cos(0.8)',Cos(0.8));
  IntTestExpressionRealResult('tan(0.8)',Tan(0.8));

  IntTestExpressionRealResult('arcsin(0.8)',ArcSin(0.8));
  IntTestExpressionRealResult('arccos(0.8)',ArcCos(0.8));
  IntTestExpressionRealResult('arctan(0.8)',ArcTan(0.8));

  IntTestExpressionRealResult('sinh(0.8)',SinH(0.8));
  IntTestExpressionRealResult('cosh(0.8)',CosH(0.8));
  IntTestExpressionRealResult('tanh(0.8)',TanH(0.8));

  IntTestExpressionRealResult('arcsinh(0.8)',ArcSinH(0.8));
  IntTestExpressionRealResult('arccosh(1.8)',ArcCosH(1.8));
  IntTestExpressionRealResult('arctanh(0.8)',ArcTanH(0.8));

  IntTestExpressionRealResult('power(0.8,0.7)',Power(0.8,0.7));
  IntTestExpressionRealResult('ArcTan2(0.8,0.7)',ArcTan2(0.8,0.7));

end;

procedure TTestExpressions.IntTestExpressionBooleanResult(const AnExpression: string; AnExpectedResult: Boolean);
var
  Value: TEvaluationValue;
  ErrMessage: string;
  Bool: Boolean;
begin
  Value := TestExpression(AnExpression, evkBoolean);
  try
    Check(Value.TryGetAsBoolean(Bool, ErrMessage), Format('Failed to convert evaluation-result of expression [%s] to boolean: [%s]', [AnExpression, ErrMessage]));
    CheckEquals(AnExpectedResult, Bool, Format('Result of expression [%s]', [AnExpression]));
  finally
    Value.Free;
  end;
end;

initialization
  RegisterTest(TTestExpressions);
end.

